"""Product block for {class}`Router` products."""
import ipaddress
from typing import Optional

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle, strEnum

from gso.products.product_blocks.site import SiteBlock, SiteBlockInactive, SiteBlockProvisioning


class RouterVendor(strEnum):
    """Enumerator for the different product vendors that are supported."""

    JUNIPER = "juniper"
    """Juniper routers."""
    NOKIA = "nokia"
    """Nokia routers."""


class RouterRole(strEnum):
    """Enumerator for the different types of routers."""

    P = "p"
    """P router."""
    PE = "pe"
    """PE router."""
    AMT = "amt"
    """AMT router."""


class RouterBlockInactive(
    ProductBlockModel, lifecycle=[SubscriptionLifecycle.INITIAL], product_block_name="RouterBlock"
):
    """A router that's being currently inactive. See {class}`RouterBlock`."""

    router_fqdn: Optional[str] = None
    router_ts_port: Optional[int] = None
    router_access_via_ts: Optional[bool] = None
    router_lo_ipv4_address: Optional[ipaddress.IPv4Address] = None
    router_lo_ipv6_address: Optional[ipaddress.IPv6Address] = None
    router_lo_iso_address: Optional[str] = None
    router_si_ipv4_network: Optional[ipaddress.IPv4Network] = None
    router_ias_lt_ipv4_network: Optional[ipaddress.IPv4Network] = None
    router_ias_lt_ipv6_network: Optional[ipaddress.IPv6Network] = None
    router_vendor: Optional[RouterVendor] = None
    router_role: Optional[RouterRole] = None
    router_site: Optional[SiteBlockInactive]
    router_is_ias_connected: Optional[bool] = None


class RouterBlockProvisioning(RouterBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """A router that's being provisioned. See {class}`RouterBlock`."""

    router_fqdn: str
    router_ts_port: int
    router_access_via_ts: Optional[bool] = None
    router_lo_ipv4_address: Optional[ipaddress.IPv4Address] = None
    router_lo_ipv6_address: Optional[ipaddress.IPv6Address] = None
    router_lo_iso_address: Optional[str] = None
    router_si_ipv4_network: Optional[ipaddress.IPv4Network] = None
    router_ias_lt_ipv4_network: Optional[ipaddress.IPv4Network] = None
    router_ias_lt_ipv6_network: Optional[ipaddress.IPv6Network] = None
    router_vendor: Optional[RouterVendor] = None
    router_role: Optional[RouterRole] = None
    router_site: Optional[SiteBlockProvisioning]
    router_is_ias_connected: Optional[bool] = None


class RouterBlock(RouterBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """A router that's currently deployed in the network."""

    router_fqdn: str
    """{term}`FQDN` of a router."""
    router_ts_port: int
    """The port of the terminal server that this router is connected to. Used for the same reason as mentioned
    previously."""
    router_access_via_ts: bool
    """Whether this router should be accessed through the terminal server, or through its loopback address."""
    router_lo_ipv4_address: ipaddress.IPv4Address
    """The IPv4 loopback address of the router."""
    router_lo_ipv6_address: ipaddress.IPv6Address
    """The IPv6 loopback address of the router."""
    router_lo_iso_address: str
    """The {term}`ISO` {term}`NET` of the router, used for {term}`IS-IS` support."""
    router_si_ipv4_network: Optional[ipaddress.IPv4Network]
    """The SI IPv4 network of the router."""
    router_ias_lt_ipv4_network: Optional[ipaddress.IPv4Network]
    """The IAS LT IPv4 network of the router."""
    router_ias_lt_ipv6_network: Optional[ipaddress.IPv6Network]
    """The IAS LT IPv6 network of the router."""
    router_vendor: RouterVendor
    """The vendor of the router, can be any of the values defined in {class}`RouterVendor`."""
    router_role: RouterRole
    """The role of the router, which can be any of the values defined in {class}`RouterRole`."""
    router_site: SiteBlock
    """The {class}`Site` that this router resides in. Both physically and computationally."""
    router_is_ias_connected: bool
    """The router is going to have a LT interface between inet0 and IAS"""
