from orchestrator.db.models import ProductTable, SubscriptionTable
from orchestrator.forms import FormPage
from orchestrator.forms.validators import Choice, UniqueConstrainedList
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import StepList, done, init, step, workflow
from orchestrator.workflows.steps import resync, set_status, store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form

from gso.products.product_blocks import PhyPortCapacity
from gso.products.product_blocks.iptrunk import IptrunkType
from gso.products.product_types.iptrunk import IptrunkInactive, IptrunkProvisioning
from gso.products.product_types.router import Router
from gso.services import ipam, provisioning_proxy, subscriptions
from gso.services.provisioning_proxy import pp_interaction
from gso.workflows.utils import customer_selector


def initial_input_form_generator(product_name: str) -> FormGenerator:
    # TODO: implement more strict validation:
    # * interface names must be validated

    routers = {}
    for router_id, router_description in subscriptions.get_active_router_subscriptions(
        fields=["subscription_id", "description"]
    ):
        routers[str(router_id)] = router_description

    class CreateIptrunkForm(FormPage):
        class Config:
            title = product_name

        customer: customer_selector()  # type: ignore
        geant_s_sid: str
        iptrunk_description: str
        iptrunk_type: IptrunkType
        iptrunk_speed: PhyPortCapacity
        iptrunk_minimum_links: int

    initial_user_input = yield CreateIptrunkForm

    class AeMembersListA(UniqueConstrainedList[str]):
        min_items = initial_user_input.iptrunk_minimum_links

    RouterEnumA = Choice("Select a router", zip(routers.keys(), routers.items()))  # type: ignore

    class CreateIptrunkSideAForm(FormPage):
        class Config:
            title = "Provide subscription details for side A of the trunk."

        iptrunk_sideA_node_id: RouterEnumA  # type: ignore
        iptrunk_sideA_ae_iface: str
        iptrunk_sideA_ae_geant_a_sid: str
        iptrunk_sideA_ae_members: AeMembersListA
        iptrunk_sideA_ae_members_descriptions: AeMembersListA

    user_input_side_a = yield CreateIptrunkSideAForm

    # Remove the selected router for side A, to prevent any loops
    routers.pop(str(user_input_side_a.iptrunk_sideA_node_id.name))
    RouterEnumB = Choice("Select a router", zip(routers.keys(), routers.items()))  # type: ignore

    class AeMembersListB(UniqueConstrainedList[str]):
        min_items = len(user_input_side_a.iptrunk_sideA_ae_members)
        max_items = len(user_input_side_a.iptrunk_sideA_ae_members)

    class CreateIptrunkSideBForm(FormPage):
        class Config:
            title = "Provide subscription details for side B of the trunk."

        iptrunk_sideB_node_id: RouterEnumB  # type: ignore
        iptrunk_sideB_ae_iface: str
        iptrunk_sideB_ae_geant_a_sid: str
        iptrunk_sideB_ae_members: AeMembersListB
        iptrunk_sideB_ae_members_descriptions: AeMembersListB

    user_input_side_b = yield CreateIptrunkSideBForm

    return initial_user_input.dict() | user_input_side_a.dict() | user_input_side_b.dict()


@step("Create subscription")
def create_subscription(product: UUIDstr, customer: UUIDstr) -> State:
    subscription = IptrunkInactive.from_product_id(product, customer)

    return {
        "subscription": subscription,
        "subscription_id": subscription.subscription_id,
    }


@step("Get information from IPAM")
def get_info_from_ipam(subscription: IptrunkProvisioning) -> State:
    # TODO: get info about how these should be generated
    subscription.iptrunk.iptrunk_ipv4_network = ipam.allocate_ipv4_network(
        service_type="TRUNK",
        comment=subscription.iptrunk.iptrunk_description,
    ).v4
    subscription.iptrunk.iptrunk_ipv6_network = ipam.allocate_ipv6_network(
        service_type="TRUNK",
        comment=subscription.iptrunk.iptrunk_description,
    ).v6
    return {"subscription": subscription}


@step("Initialize subscription")
def initialize_subscription(
    subscription: IptrunkInactive,
    geant_s_sid: str,
    iptrunk_type: IptrunkType,
    iptrunk_description: str,
    iptrunk_speed: str,
    iptrunk_minimum_links: int,
    iptrunk_sideA_node_id: str,
    iptrunk_sideA_ae_iface: str,
    iptrunk_sideA_ae_geant_a_sid: str,
    iptrunk_sideA_ae_members: list[str],
    iptrunk_sideA_ae_members_descriptions: list[str],
    iptrunk_sideB_node_id: str,
    iptrunk_sideB_ae_iface: str,
    iptrunk_sideB_ae_geant_a_sid: str,
    iptrunk_sideB_ae_members: list[str],
    iptrunk_sideB_ae_members_descriptions: list[str],
) -> State:
    subscription.iptrunk.geant_s_sid = geant_s_sid
    subscription.iptrunk.iptrunk_description = iptrunk_description
    subscription.iptrunk.iptrunk_type = iptrunk_type
    subscription.iptrunk.iptrunk_speed = iptrunk_speed
    subscription.iptrunk.iptrunk_isis_metric = 9000
    subscription.iptrunk.iptrunk_minimum_links = iptrunk_minimum_links

    subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node = Router.from_subscription(iptrunk_sideA_node_id).router
    subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_iface = iptrunk_sideA_ae_iface
    subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_geant_a_sid = iptrunk_sideA_ae_geant_a_sid
    subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members = iptrunk_sideA_ae_members
    subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members_description = iptrunk_sideA_ae_members_descriptions

    subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node = Router.from_subscription(iptrunk_sideB_node_id).router
    subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_iface = iptrunk_sideB_ae_iface
    subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_geant_a_sid = iptrunk_sideB_ae_geant_a_sid
    subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_members = iptrunk_sideB_ae_members
    subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_members_description = iptrunk_sideB_ae_members_descriptions

    subscription.description = f"IP trunk, geant_s_sid:{geant_s_sid}"
    subscription = IptrunkProvisioning.from_other_lifecycle(subscription, SubscriptionLifecycle.PROVISIONING)

    return {"subscription": subscription}


@step("Provision IP trunk interface [DRY RUN]")
def provision_ip_trunk_iface_dry(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "trunk_interface")

    return {
        "subscription": subscription,
        "label_text": "[DRY RUN] Provisioning a trunk interface, please refresh to get the results of the playbook.",
    }


@step("Provision IP trunk interface [FOR REAL]")
def provision_ip_trunk_iface_real(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "trunk_interface", False)

    return {
        "subscription": subscription,
        "label_text": "Provisioning a trunk interface, please refresh to get the results of the playbook.",
    }


@step("Check IP connectivity of the trunk")
def check_ip_trunk_connectivity(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.check_ip_trunk(subscription, process_id, "ping")

    return {
        "subscription": subscription,
        "label_text": "Checking IP traffic flow on the trunk, to get the results of the playbook.",
    }


@step("Provision IP trunk ISIS interface [DRY RUN]")
def provision_ip_trunk_isis_iface_dry(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "isis_interface")

    return {
        "subscription": subscription,
        "label_text": "[DRY RUN] Provisioning ISIS interfaces, please refresh to get the results of the playbook.",
    }


@step("Provision IP trunk ISIS interface [FOR REAL]")
def provision_ip_trunk_isis_iface_real(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "isis_interface", False)

    return {
        "subscription": subscription,
        "label_text": "Provisioning ISIS interfaces, please refresh to get the results of the playbook.",
    }


@step("Check ISIS adjacency")
def check_ip_trunk_isis(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.check_ip_trunk(subscription, process_id, "isis")

    return {
        "subscription": subscription,
        "label_text": "Checking ISIS adjacency, please refresh to get the results of the playbook.",
    }


@step("Provision IP trunk LDP interface [DRY RUN]")
def provision_ip_trunk_ldp_iface_dry(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "ldp_interface")

    return {
        "subscription": subscription,
        "label_text": "[DRY RUN] Provisioning LDP interface, please refresh to get the results of the playbook.",
    }


@step("Provision IP trunk LDP interface [FOR REAL]")
def provision_ip_trunk_ldp_iface_real(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "ldp_interface", False)

    return {
        "subscription": subscription,
        "label_text": "Provisioning LDP interface, please refresh to get the results of the playbook.",
    }


@workflow(
    "Create IP trunk",
    initial_input_form=wrap_create_initial_input_form(initial_input_form_generator),
    target=Target.CREATE,
)
def create_iptrunk() -> StepList:
    return (
        init
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> get_info_from_ipam
        >> pp_interaction(provision_ip_trunk_iface_dry, 3)
        >> pp_interaction(provision_ip_trunk_iface_real, 3)
        >> pp_interaction(check_ip_trunk_connectivity, 2, False)
        >> pp_interaction(provision_ip_trunk_isis_iface_dry, 3)
        >> pp_interaction(provision_ip_trunk_isis_iface_real, 3)
        >> pp_interaction(check_ip_trunk_isis, 2, False)
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
